import datetime
import re
import typing

import betterproto2

from betterproto2_compiler.lib.google.protobuf import Duration as VanillaDuration


class Duration(VanillaDuration):
    @classmethod
    def from_timedelta(
        cls, delta: datetime.timedelta, *, _1_microsecond: datetime.timedelta = datetime.timedelta(microseconds=1)
    ) -> "Duration":
        total_ms = delta // _1_microsecond
        seconds = int(total_ms / 1e6)
        nanos = int((total_ms % 1e6) * 1e3)
        return cls(seconds, nanos)

    def to_timedelta(self) -> datetime.timedelta:
        return datetime.timedelta(seconds=self.seconds, microseconds=self.nanos / 1e3)

    @staticmethod
    def delta_to_json(delta: datetime.timedelta) -> str:
        parts = str(delta.total_seconds()).split(".")
        if len(parts) > 1:
            while len(parts[1]) not in (3, 6, 9):
                parts[1] = f"{parts[1]}0"
        return f"{'.'.join(parts)}s"

    # TODO typing
    @classmethod
    def from_dict(cls, value, *, ignore_unknown_fields: bool = False):
        if isinstance(value, str):
            if not re.match(r"^\d+(\.\d+)?s$", value):
                raise ValueError(f"Invalid duration string: {value}")

            seconds = float(value[:-1])
            return Duration(seconds=int(seconds), nanos=int((seconds - int(seconds)) * 1e9))

        return super().from_dict(value, ignore_unknown_fields=ignore_unknown_fields)

    # TODO typing
    def to_dict(
        self,
        *,
        output_format: betterproto2.OutputFormat = betterproto2.OutputFormat.PROTO_JSON,
        casing: betterproto2.Casing = betterproto2.Casing.CAMEL,
        include_default_values: bool = False,
    ) -> dict[str, typing.Any] | typing.Any:
        # If the output format is PYTHON, we should have kept the wrapped type without building the real class
        assert output_format == betterproto2.OutputFormat.PROTO_JSON

        assert 0 <= self.nanos < 1e9

        if self.nanos == 0:
            return f"{self.seconds}s"

        nanos = f"{self.nanos:09d}".rstrip("0")
        if len(nanos) < 3:
            nanos += "0" * (3 - len(nanos))

        return f"{self.seconds}.{nanos}s"

    @staticmethod
    def from_wrapped(wrapped: datetime.timedelta) -> "Duration":
        return Duration.from_timedelta(wrapped)

    def to_wrapped(self) -> datetime.timedelta:
        return self.to_timedelta()



"""Plugin model dataclasses.

These classes are meant to be an intermediate representation
of protobuf objects. They are used to organize the data collected during parsing.

The general intention is to create a doubly-linked tree-like structure
with the following types of references:
- Downwards references: from message -> fields, from output package -> messages
or from service -> service methods
- Upwards references: from field -> message, message -> package.
- Input/output message references: from a service method to it's corresponding
input/output messages, which may even be in another package.

There are convenience methods to allow climbing up and down this tree, for
example to retrieve the list of all messages that are in the same package as
the current message.

Most of these classes take as inputs:
- proto_obj: A reference to it's corresponding protobuf object as
presented by the protoc plugin.
- parent: a reference to the parent object in the tree.

With this information, the class is able to expose attributes,
such as a pythonized name, that will be calculated from proto_obj.
"""

import builtins
import inspect
from collections.abc import Iterator
from dataclasses import (
    dataclass,
    field,
)

from betterproto2 import unwrap

from betterproto2_compiler import casing
from betterproto2_compiler.compile.importing import get_type_reference, parse_source_type_name
from betterproto2_compiler.compile.naming import (
    pythonize_class_name,
    pythonize_field_name,
    pythonize_method_name,
)
from betterproto2_compiler.known_types import KNOWN_METHODS, WRAPPED_TYPES
from betterproto2_compiler.lib.google.protobuf import (
    DescriptorProto,
    EnumDescriptorProto,
    FieldDescriptorProto,
    FieldDescriptorProtoLabel,
    FieldDescriptorProtoType,
    FieldDescriptorProtoType as FieldType,
    FileDescriptorProto,
    MethodDescriptorProto,
    OneofDescriptorProto,
    ServiceDescriptorProto,
    SourceCodeInfo,
)
from betterproto2_compiler.lib.google.protobuf.compiler import CodeGeneratorRequest
from betterproto2_compiler.settings import Settings

# Organize proto types into categories
PROTO_FLOAT_TYPES = (
    FieldDescriptorProtoType.DOUBLE,  # 1
    FieldDescriptorProtoType.FLOAT,  # 2
)
PROTO_INT_TYPES = (
    FieldDescriptorProtoType.INT64,  # 3
    FieldDescriptorProtoType.UINT64,  # 4
    FieldDescriptorProtoType.INT32,  # 5
    FieldDescriptorProtoType.FIXED64,  # 6
    FieldDescriptorProtoType.FIXED32,  # 7
    FieldDescriptorProtoType.UINT32,  # 13
    FieldDescriptorProtoType.SFIXED32,  # 15
    FieldDescriptorProtoType.SFIXED64,  # 16
    FieldDescriptorProtoType.SINT32,  # 17
    FieldDescriptorProtoType.SINT64,  # 18
)
PROTO_BOOL_TYPES = (FieldDescriptorProtoType.BOOL,)  # 8
PROTO_STR_TYPES = (FieldDescriptorProtoType.STRING,)  # 9
PROTO_BYTES_TYPES = (FieldDescriptorProtoType.BYTES,)  # 12
PROTO_MESSAGE_TYPES = (
    FieldDescriptorProtoType.MESSAGE,  # 11
    FieldDescriptorProtoType.ENUM,  # 14
)
PROTO_MAP_TYPES = (FieldDescriptorProtoType.MESSAGE,)  # 11
PROTO_PACKED_TYPES = (
    FieldDescriptorProtoType.DOUBLE,  # 1
    FieldDescriptorProtoType.FLOAT,  # 2
    FieldDescriptorProtoType.INT64,  # 3
    FieldDescriptorProtoType.UINT64,  # 4
    FieldDescriptorProtoType.INT32,  # 5
    FieldDescriptorProtoType.FIXED64,  # 6
    FieldDescriptorProtoType.FIXED32,  # 7
    FieldDescriptorProtoType.BOOL,  # 8
    FieldDescriptorProtoType.UINT32,  # 13
    FieldDescriptorProtoType.SFIXED32,  # 15
    FieldDescriptorProtoType.SFIXED64,  # 16
    FieldDescriptorProtoType.SINT32,  # 17
    FieldDescriptorProtoType.SINT64,  # 18
)


def get_comment(
    proto_file: "FileDescriptorProto",
    path: list[int],
) -> str:
    for sci_loc in unwrap(proto_file.source_code_info).location:
        if list(sci_loc.path) == path:
            all_comments = list(sci_loc.leading_detached_comments)
            if sci_loc.leading_comments:
                all_comments.append(sci_loc.leading_comments)
            if sci_loc.trailing_comments:
                all_comments.append(sci_loc.trailing_comments)

            lines = []

            for comment in all_comments:
                lines += comment.split("\n")
                lines.append("")

            # Remove consecutive empty lines
            lines = [line for i, line in enumerate(lines) if line or (i == 0 or lines[i - 1])]

            if lines and not lines[-1]:
                lines.pop()  # Remove the last empty line

            # It is common for one line comments to start with a space, for example: // comment
            # We don't add this space to the generated file.
            lines = [line[1:] if line and line[0] == " " else line for line in lines]

            comment = "\n".join(lines)

            # Escape backslashes and triple quotes
            comment = comment.replace("\\", "\\\\").replace('"""', '\\"\\"\\"')

            return comment

    return ""


@dataclass(kw_only=True)
class ProtoContentBase:
    """Methods common to MessageCompiler, ServiceCompiler and ServiceMethodCompiler."""

    source_file: FileDescriptorProto
    path: list[int]

    def ready(self) -> None:
        """
        This function is called after all the compilers are created, but before generating the output code.
        """
        pass

    @property
    def comment(self) -> str:
        """Crawl the proto source code and retrieve comments
        for this object.
        """
        return get_comment(proto_file=self.source_file, path=self.path)


@dataclass(kw_only=True)
class PluginRequestCompiler:
    plugin_request_obj: CodeGeneratorRequest
    output_packages: dict[str, "OutputTemplate"] = field(default_factory=dict)

    @property
    def all_messages(self) -> list["MessageCompiler"]:
        """All of the messages in this request.

        Returns
        -------
        List[MessageCompiler]
            List of all of the messages in this request.
        """
        return [msg for output in self.output_packages.values() for msg in output.messages.values()]


@dataclass(kw_only=True)
class OutputTemplate:
    """Representation of an output .py file.

    Each output file corresponds to a .proto input file,
    but may need references to other .proto files to be
    built.
    """

    parent_request: PluginRequestCompiler
    package_proto_obj: FileDescriptorProto
    input_files: list[FileDescriptorProto] = field(default_factory=list)
    imports_end: set[str] = field(default_factory=set)
    messages: dict[str, "MessageCompiler"] = field(default_factory=dict)
    enums: dict[str, "EnumDefinitionCompiler"] = field(default_factory=dict)
    services: dict[str, "ServiceCompiler"] = field(default_factory=dict)

    settings: Settings

    @property
    def package(self) -> str:
        """Name of input package.

        Returns
        -------
        str
            Name of input package.
        """
        return self.package_proto_obj.package

    @property
    def input_filenames(self) -> list[str]:
        """Names of the input files used to build this output.

        Returns
        -------
        list[str]
            Names of the input files used to build this output.
        """
        return sorted([f.name for f in self.input_files])

    def get_descriptor_name(self, source_file: FileDescriptorProto):
        return f"{source_file.name.replace('/', '_').replace('.', '_').upper()}_DESCRIPTOR"

    @property
    def descriptors(self):
        """Google protobuf library descriptors.

        Returns
        -------
        str
            A list of pool registrations for proto descriptors.
        """
        descriptors: list[str] = []

        for f in self.input_files:
            # Remove the source_code_info field since it is not needed at runtime.
            source_code_info: SourceCodeInfo | None = f.source_code_info
            f.source_code_info = None

            descriptors.append(
                f"{self.get_descriptor_name(f)} = default_google_proto_descriptor_pool.AddSerializedFile({bytes(f)})"
            )

            f.source_code_info = source_code_info

        return "\n".join(descriptors)


@dataclass(kw_only=True)
class MessageCompiler(ProtoContentBase):
    """Representation of a protobuf message."""

    output_file: OutputTemplate
    proto_obj: DescriptorProto
    prefixed_proto_name: str
    fields: list["FieldCompiler"] = field(default_factory=list)
    oneofs: list["OneofCompiler"] = field(default_factory=list)
    builtins_types: set[str] = field(default_factory=set)

    @property
    def proto_name(self) -> str:
        return self.proto_obj.name

    @property
    def py_name(self) -> str:
        return pythonize_class_name(self.prefixed_proto_name)

    @property
    def deprecated(self) -> bool:
        return bool(self.proto_obj.options and self.proto_obj.options.deprecated)

    @property
    def deprecated_fields(self) -> Iterator[str]:
        for f in self.fields:
            if f.deprecated:
                yield f.py_name

    @property
    def has_deprecated_fields(self) -> bool:
        return any(self.deprecated_fields)

    @property
    def has_oneof_fields(self) -> bool:
        return any(isinstance(field, OneOfFieldCompiler) for field in self.fields)

    @property
    def custom_methods(self) -> list[str]:
        """
        Return a list of the custom methods.
        """
        methods_source: list[str] = []

        for method in KNOWN_METHODS.get((self.source_file.package, self.py_name), []):
            source = inspect.getsource(method)
            methods_source.append(source.strip())

        return methods_source

    @property
    def descriptor_name(self) -> str:
        """Google protobuf library descriptor name.

        Returns
        -------
        str
            The Python name of the descriptor to reference.
        """
        return self.output_file.get_descriptor_name(self.source_file)


def is_map(proto_field_obj: FieldDescriptorProto, parent_message: DescriptorProto) -> bool:
    """True if proto_field_obj is a map, otherwise False."""
    if proto_field_obj.type == FieldDescriptorProtoType.MESSAGE:
        if not hasattr(parent_message, "nested_type"):
            return False

        # This might be a map...
        message_type = proto_field_obj.type_name.split(".").pop().lower()
        map_entry = f"{proto_field_obj.name.replace('_', '').lower()}entry"
        if message_type == map_entry:
            for nested in parent_message.nested_type:  # parent message
                if nested.name.replace("_", "").lower() == map_entry and nested.options and nested.options.map_entry:
                    return True
    return False


def is_oneof(proto_field_obj: FieldDescriptorProto) -> bool:
    """
    True if proto_field_obj is a OneOf, otherwise False.
    """
    return not proto_field_obj.proto3_optional and proto_field_obj.oneof_index is not None


@dataclass(kw_only=True)
class FieldCompiler(ProtoContentBase):
    builtins_types: set[str] = field(default_factory=set)

    message: MessageCompiler
    proto_obj: FieldDescriptorProto

    @property
    def output_file(self) -> "OutputTemplate":
        return self.message.output_file

    def get_field_string(self) -> str:
        """Construct string representation of this field as a field."""
        name = f"{self.py_name}"
        field_args = ", ".join(([""] + self.betterproto_field_args) if self.betterproto_field_args else [])

        betterproto_field_type = (
            f"betterproto2.field({self.proto_obj.number}, betterproto2.TYPE_{str(self.field_type)}{field_args})"
        )
        if self.py_name in dir(builtins):
            self.message.builtins_types.add(self.py_name)
        return f'{name}: "{self.annotation}" = {betterproto_field_type}'

    @property
    def betterproto_field_args(self) -> list[str]:
        args = []

        if self.field_type == FieldDescriptorProtoType.MESSAGE and self.is_wrapped:
            unwrap_type = self.unwrapped_py_type

            # Without the lambda function, the type is evaluated right away, which fails since the corresponding
            # import is placed at the end of the file to avoid circular imports.
            args.append(f"unwrap=lambda: {unwrap_type}")

        if self.optional:
            args.append("optional=True")
        elif self.repeated:
            args.append("repeated=True")
        elif self.field_type == FieldType.ENUM:
            args.append(f"default_factory=lambda: {self.py_type}(0)")
        return args

    @property
    def deprecated(self) -> bool:
        return bool(self.proto_obj.options and self.proto_obj.options.deprecated)

    @property
    def use_builtins(self) -> bool:
        return self.py_type in self.message.builtins_types or (
            self.py_type == self.py_name and self.py_name in dir(builtins)
        )

    @property
    def repeated(self) -> bool:
        return self.proto_obj.label == FieldDescriptorProtoLabel.REPEATED

    @property
    def optional(self) -> bool:
        # TODO not for maps
        return self.proto_obj.proto3_optional or (self.field_type == FieldType.MESSAGE and not self.repeated)

    @property
    def field_type(self) -> FieldType:
        return self.proto_obj.type

    @property
    def packed(self) -> bool:
        """True if the wire representation is a packed format."""
        return self.repeated and self.proto_obj.type in PROTO_PACKED_TYPES

    @property
    def py_name(self) -> str:
        """Pythonized name."""
        return pythonize_field_name(self.proto_name)

    @property
    def proto_name(self) -> str:
        """Original protobuf name."""
        return self.proto_obj.name

    @property
    def is_wrapped(self) -> bool:
        assert self.field_type == FieldDescriptorProtoType.MESSAGE
        type_package, type_name = parse_source_type_name(self.proto_obj.type_name, self.output_file.parent_request)

        return (type_package, type_name) in WRAPPED_TYPES

    def _py_type(self, wrap: bool) -> str:
        """String representation of Python type."""
        if self.proto_obj.type in PROTO_FLOAT_TYPES:
            return "float"
        elif self.proto_obj.type in PROTO_INT_TYPES:
            return "int"
        elif self.proto_obj.type in PROTO_BOOL_TYPES:
            return "bool"
        elif self.proto_obj.type in PROTO_STR_TYPES:
            return "str"
        elif self.proto_obj.type in PROTO_BYTES_TYPES:
            return "bytes"
        elif self.proto_obj.type in PROTO_MESSAGE_TYPES:
            # Type referencing another defined Message or a named enum
            return get_type_reference(
                package=self.output_file.package,
                imports=self.output_file.imports_end,
                source_type=self.proto_obj.type_name,
                request=self.output_file.parent_request,
                wrap=wrap,
                settings=self.output_file.settings,
            )
        else:
            raise NotImplementedError(f"Unknown type {self.proto_obj.type}")

    @property
    def py_type(self) -> str:
        return self._py_type(wrap=True)

    @property
    def unwrapped_py_type(self) -> str:
        return self._py_type(wrap=False)

    @property
    def annotations(self) -> list[str]:
        """List of the Pydantic annotation to add to the field."""
        assert self.output_file.settings.pydantic_dataclasses

        annotations = []

        if self.proto_obj.type in (FieldType.INT32, FieldType.SFIXED32, FieldType.SINT32):
            annotations.append("pydantic.Field(ge=-2**31, le=2**31 - 1)")

        elif self.proto_obj.type in (FieldType.UINT32, FieldType.FIXED32):
            annotations.append("pydantic.Field(ge=0, le=2**32 - 1)")

        elif self.proto_obj.type in (FieldType.INT64, FieldType.SFIXED64, FieldType.SINT64):
            annotations.append("pydantic.Field(ge=-2**63, le=2**63 - 1)")

        elif self.proto_obj.type in (FieldType.UINT64, FieldType.FIXED64):
            annotations.append("pydantic.Field(ge=0, le=2**64 - 1)")

        elif self.proto_obj.type == FieldType.FLOAT:
            annotations.append("pydantic.AfterValidator(betterproto2.validators.validate_float32)")

        elif self.proto_obj.type == FieldType.STRING:
            annotations.append("pydantic.AfterValidator(betterproto2.validators.validate_string)")

        return annotations

    @property
    def annotation(self) -> str:
        py_type = self.py_type

        if self.use_builtins:
            py_type = f"builtins.{py_type}"

        # Add the pydantic annotation if needed
        if self.output_file.settings.pydantic_dataclasses:
            annotations = self.annotations
            if annotations:
                py_type = f"typing.Annotated[{py_type}, {', '.join(annotations)}]"

        if self.repeated:
            return f"list[{py_type}]"
        if self.optional:
            return f"{py_type} | None"
        return py_type


@dataclass(kw_only=True)
class OneOfFieldCompiler(FieldCompiler):
    @property
    def optional(self) -> bool:
        return True

    @property
    def betterproto_field_args(self) -> list[str]:
        args = super().betterproto_field_args
        group = self.message.proto_obj.oneof_decl[unwrap(self.proto_obj.oneof_index)].name
        args.append(f'group="{group}"')
        return args


@dataclass(kw_only=True)
class MapEntryCompiler(FieldCompiler):
    py_k_type: str = ""
    py_v_type: str = ""
    unwrap_v: str | None = None

    proto_k_type: str = ""
    proto_v_type: str = ""

    def ready(self) -> None:
        """Explore nested types and set k_type and v_type if unset."""
        map_entry = f"{self.proto_obj.name.replace('_', '').lower()}entry"
        for nested in self.message.proto_obj.nested_type:
            if nested.name.replace("_", "").lower() == map_entry and unwrap(nested.options).map_entry:
                # Get Python types
                assert nested.field[0].name == "key"
                self.py_k_type = FieldCompiler(
                    source_file=self.source_file,
                    proto_obj=nested.field[0],  # key
                    path=[],
                    message=self.message,
                ).py_type

                assert nested.field[1].name == "value"
                value_field_compiler = FieldCompiler(
                    source_file=self.source_file,
                    proto_obj=nested.field[1],  # value
                    path=[],
                    message=self.message,
                )

                self.py_v_type = value_field_compiler.py_type
                if (
                    value_field_compiler.field_type == FieldDescriptorProtoType.MESSAGE
                    and value_field_compiler.is_wrapped
                ):
                    self.unwrap_v = value_field_compiler.unwrapped_py_type
                else:
                    self.unwrap_v = None

                # Get proto types
                self.proto_k_type = unwrap(FieldDescriptorProtoType(nested.field[0].type).name)
                self.proto_v_type = unwrap(FieldDescriptorProtoType(nested.field[1].type).name)
                return

        raise ValueError("can't find enum")

    def get_field_string(self) -> str:
        """Construct string representation of this field as a field."""
        proto_type_1 = f"betterproto2.TYPE_{self.proto_k_type}"
        proto_type_2 = f"betterproto2.TYPE_{self.proto_v_type}"

        unwrap_2 = ""
        if self.unwrap_v:
            unwrap_2 = f", unwrap_2=lambda: {self.unwrap_v}"

        betterproto_field_type = (
            "betterproto2.field("
            f"{self.proto_obj.number}, "
            "betterproto2.TYPE_MAP, "
            f"map_meta=betterproto2.map_meta({proto_type_1}, {proto_type_2}{unwrap_2})"
            ")"
        )
        if self.py_name in dir(builtins):
            self.message.builtins_types.add(self.py_name)
        return f'{self.py_name}: "{self.annotation}" = {betterproto_field_type}'

    @property
    def annotation(self) -> str:
        return f"dict[{self.py_k_type}, {self.py_v_type}]"

    @property
    def repeated(self) -> bool:
        return False  # maps cannot be repeated


@dataclass(kw_only=True)
class OneofCompiler(ProtoContentBase):
    proto_obj: OneofDescriptorProto

    @property
    def name(self) -> str:
        return self.proto_obj.name


@dataclass(kw_only=True)
class EnumDefinitionCompiler(ProtoContentBase):
    """Representation of a proto Enum definition."""

    output_file: OutputTemplate
    proto_obj: EnumDescriptorProto
    prefixed_proto_name: str
    entries: list["EnumDefinitionCompiler.EnumEntry"] = field(default_factory=list)

    @dataclass(unsafe_hash=True, kw_only=True)
    class EnumEntry:
        """Representation of an Enum entry."""

        name: str
        proto_name: str
        value: int
        comment: str

    def __post_init__(self) -> None:
        self.entries = [
            self.EnumEntry(
                name=entry_proto_value.name,
                proto_name=entry_proto_value.name,
                value=entry_proto_value.number,
                comment=get_comment(proto_file=self.source_file, path=self.path + [2, entry_number]),
            )
            for entry_number, entry_proto_value in enumerate(self.proto_obj.value)
        ]

        if not self.entries:
            return

        # Remove enum prefixes
        enum_name: str = self.proto_obj.name

        enum_name_reduced = enum_name.replace("_", "").lower()

        first_entry = self.entries[0].name

        # Find the potential common prefix
        enum_prefix = ""
        for i in range(len(first_entry)):
            if first_entry[: i + 1].replace("_", "").lower() == enum_name_reduced:
                enum_prefix = f"{first_entry[: i + 1]}_"
                break

        should_rename = enum_prefix and all(entry.name.startswith(enum_prefix) for entry in self.entries)

        if should_rename:
            for entry in self.entries:
                entry.name = entry.name[len(enum_prefix) :]

        for entry in self.entries:
            entry.name = casing.sanitize_name(entry.name)

    @property
    def proto_name(self) -> str:
        return self.proto_obj.name

    @property
    def py_name(self) -> str:
        return pythonize_class_name(self.prefixed_proto_name)

    @property
    def deprecated(self) -> bool:
        return bool(self.proto_obj.options and self.proto_obj.options.deprecated)

    @property
    def descriptor_name(self) -> str:
        """Google protobuf library descriptor name.

        Returns
        -------
        str
            The Python name of the descriptor to reference.
        """
        return self.output_file.get_descriptor_name(self.source_file)

    @property
    def has_renamed_entries(self) -> bool:
        return any(entry.proto_name != entry.name for entry in self.entries)


@dataclass(kw_only=True)
class ServiceCompiler(ProtoContentBase):
    output_file: OutputTemplate
    proto_obj: ServiceDescriptorProto
    methods: list["ServiceMethodCompiler"] = field(default_factory=list)

    @property
    def proto_name(self) -> str:
        return self.proto_obj.name

    @property
    def py_name(self) -> str:
        return pythonize_class_name(self.proto_name)


@dataclass(kw_only=True)
class ServiceMethodCompiler(ProtoContentBase):
    parent: ServiceCompiler
    proto_obj: MethodDescriptorProto

    @property
    def py_name(self) -> str:
        """Pythonized method name."""
        return pythonize_method_name(self.proto_obj.name)

    @property
    def proto_name(self) -> str:
        """Original protobuf name."""
        return self.proto_obj.name

    @property
    def route(self) -> str:
        package_part = f"{self.parent.output_file.package}." if self.parent.output_file.package else ""
        return f"/{package_part}{self.parent.proto_name}/{self.proto_name}"

    @property
    def py_input_message_type(self) -> str:
        """String representation of the Python type corresponding to the
        input message.

        Returns
        -------
        str
            String representation of the Python type corresponding to the input message.
        """
        return get_type_reference(
            package=self.parent.output_file.package,
            imports=self.parent.output_file.imports_end,
            source_type=self.proto_obj.input_type,
            request=self.parent.output_file.parent_request,
            wrap=False,
            settings=self.parent.output_file.settings,
        )

    @property
    def is_input_msg_empty(self: "ServiceMethodCompiler") -> bool:
        package, name = parse_source_type_name(self.proto_obj.input_type, self.parent.output_file.parent_request)

        msg = self.parent.output_file.parent_request.output_packages[package].messages[name]

        return not bool(msg.fields)

    @property
    def py_output_message_type(self) -> str:
        """String representation of the Python type corresponding to the
        output message.

        Returns
        -------
        str
            String representation of the Python type corresponding to the output message.
        """
        return get_type_reference(
            package=self.parent.output_file.package,
            imports=self.parent.output_file.imports_end,
            source_type=self.proto_obj.output_type,
            request=self.parent.output_file.parent_request,
            wrap=False,
            settings=self.parent.output_file.settings,
        )

    @property
    def client_streaming(self) -> bool:
        return self.proto_obj.client_streaming

    @property
    def server_streaming(self) -> bool:
        return self.proto_obj.server_streaming

    @property
    def deprecated(self) -> bool:
        return bool(self.proto_obj.options and self.proto_obj.options.deprecated)



import typing

import betterproto2

from betterproto2_compiler.lib.google.protobuf import (
    BoolValue as VanillaBoolValue,
    BytesValue as VanillaBytesValue,
    DoubleValue as VanillaDoubleValue,
    FloatValue as VanillaFloatValue,
    Int32Value as VanillaInt32Value,
    Int64Value as VanillaInt64Value,
    StringValue as VanillaStringValue,
    UInt32Value as VanillaUInt32Value,
    UInt64Value as VanillaUInt64Value,
)


class BoolValue(VanillaBoolValue):
    @staticmethod
    def from_wrapped(wrapped: bool) -> "BoolValue":
        return BoolValue(value=wrapped)

    def to_wrapped(self) -> bool:
        return self.value

    @classmethod
    def from_dict(cls, value, *, ignore_unknown_fields: bool = False):
        if isinstance(value, bool):
            return BoolValue(value=value)
        return super().from_dict(value, ignore_unknown_fields=ignore_unknown_fields)

    def to_dict(
        self,
        *,
        output_format: betterproto2.OutputFormat = betterproto2.OutputFormat.PROTO_JSON,
        casing: betterproto2.Casing = betterproto2.Casing.CAMEL,
        include_default_values: bool = False,
    ) -> dict[str, typing.Any] | typing.Any:
        return self.value


class Int32Value(VanillaInt32Value):
    @staticmethod
    def from_wrapped(wrapped: int) -> "Int32Value":
        return Int32Value(value=wrapped)

    def to_wrapped(self) -> int:
        return self.value

    @classmethod
    def from_dict(cls, value, *, ignore_unknown_fields: bool = False):
        if isinstance(value, int):
            return Int32Value(value=value)
        return super().from_dict(value, ignore_unknown_fields=ignore_unknown_fields)

    def to_dict(
        self,
        *,
        output_format: betterproto2.OutputFormat = betterproto2.OutputFormat.PROTO_JSON,
        casing: betterproto2.Casing = betterproto2.Casing.CAMEL,
        include_default_values: bool = False,
    ) -> dict[str, typing.Any] | typing.Any:
        return self.value


class Int64Value(VanillaInt64Value):
    @staticmethod
    def from_wrapped(wrapped: int) -> "Int64Value":
        return Int64Value(value=wrapped)

    def to_wrapped(self) -> int:
        return self.value

    @classmethod
    def from_dict(cls, value, *, ignore_unknown_fields: bool = False):
        if isinstance(value, int):
            return Int64Value(value=value)
        return super().from_dict(value, ignore_unknown_fields=ignore_unknown_fields)

    def to_dict(
        self,
        *,
        output_format: betterproto2.OutputFormat = betterproto2.OutputFormat.PROTO_JSON,
        casing: betterproto2.Casing = betterproto2.Casing.CAMEL,
        include_default_values: bool = False,
    ) -> dict[str, typing.Any] | typing.Any:
        return self.value


class UInt32Value(VanillaUInt32Value):
    @staticmethod
    def from_wrapped(wrapped: int) -> "UInt32Value":
        return UInt32Value(value=wrapped)

    def to_wrapped(self) -> int:
        return self.value

    @classmethod
    def from_dict(cls, value, *, ignore_unknown_fields: bool = False):
        if isinstance(value, int):
            return UInt32Value(value=value)
        return super().from_dict(value, ignore_unknown_fields=ignore_unknown_fields)

    def to_dict(
        self,
        *,
        output_format: betterproto2.OutputFormat = betterproto2.OutputFormat.PROTO_JSON,
        casing: betterproto2.Casing = betterproto2.Casing.CAMEL,
        include_default_values: bool = False,
    ) -> dict[str, typing.Any] | typing.Any:
        return self.value


class UInt64Value(VanillaUInt64Value):
    @staticmethod
    def from_wrapped(wrapped: int) -> "UInt64Value":
        return UInt64Value(value=wrapped)

    def to_wrapped(self) -> int:
        return self.value

    @classmethod
    def from_dict(cls, value, *, ignore_unknown_fields: bool = False):
        if isinstance(value, int):
            return UInt64Value(value=value)
        return super().from_dict(value, ignore_unknown_fields=ignore_unknown_fields)

    def to_dict(
        self,
        *,
        output_format: betterproto2.OutputFormat = betterproto2.OutputFormat.PROTO_JSON,
        casing: betterproto2.Casing = betterproto2.Casing.CAMEL,
        include_default_values: bool = False,
    ) -> dict[str, typing.Any] | typing.Any:
        return self.value


class FloatValue(VanillaFloatValue):
    @staticmethod
    def from_wrapped(wrapped: float) -> "FloatValue":
        return FloatValue(value=wrapped)

    def to_wrapped(self) -> float:
        return self.value

    @classmethod
    def from_dict(cls, value, *, ignore_unknown_fields: bool = False):
        if isinstance(value, float):
            return FloatValue(value=value)
        return super().from_dict(value, ignore_unknown_fields=ignore_unknown_fields)

    def to_dict(
        self,
        *,
        output_format: betterproto2.OutputFormat = betterproto2.OutputFormat.PROTO_JSON,
        casing: betterproto2.Casing = betterproto2.Casing.CAMEL,
        include_default_values: bool = False,
    ) -> dict[str, typing.Any] | typing.Any:
        return self.value


class DoubleValue(VanillaDoubleValue):
    @staticmethod
    def from_wrapped(wrapped: float) -> "DoubleValue":
        return DoubleValue(value=wrapped)

    def to_wrapped(self) -> float:
        return self.value

    @classmethod
    def from_dict(cls, value, *, ignore_unknown_fields: bool = False):
        if isinstance(value, float):
            return DoubleValue(value=value)
        return super().from_dict(value, ignore_unknown_fields=ignore_unknown_fields)

    def to_dict(
        self,
        *,
        output_format: betterproto2.OutputFormat = betterproto2.OutputFormat.PROTO_JSON,
        casing: betterproto2.Casing = betterproto2.Casing.CAMEL,
        include_default_values: bool = False,
    ) -> dict[str, typing.Any] | typing.Any:
        return self.value


class StringValue(VanillaStringValue):
    @staticmethod
    def from_wrapped(wrapped: str) -> "StringValue":
        return StringValue(value=wrapped)

    def to_wrapped(self) -> str:
        return self.value

    @classmethod
    def from_dict(cls, value, *, ignore_unknown_fields: bool = False):
        if isinstance(value, str):
            return StringValue(value=value)
        return super().from_dict(value, ignore_unknown_fields=ignore_unknown_fields)

    def to_dict(
        self,
        *,
        output_format: betterproto2.OutputFormat = betterproto2.OutputFormat.PROTO_JSON,
        casing: betterproto2.Casing = betterproto2.Casing.CAMEL,
        include_default_values: bool = False,
    ) -> dict[str, typing.Any] | typing.Any:
        return self.value


class BytesValue(VanillaBytesValue):
    @staticmethod
    def from_wrapped(wrapped: bytes) -> "BytesValue":
        return BytesValue(value=wrapped)

    def to_wrapped(self) -> bytes:
        return self.value

    @classmethod
    def from_dict(cls, value, *, ignore_unknown_fields: bool = False):
        if isinstance(value, bytes):
            return BytesValue(value=value)
        return super().from_dict(value, ignore_unknown_fields=ignore_unknown_fields)

    def to_dict(
        self,
        *,
        output_format: betterproto2.OutputFormat = betterproto2.OutputFormat.PROTO_JSON,
        casing: betterproto2.Casing = betterproto2.Casing.CAMEL,
        include_default_values: bool = False,
    ) -> dict[str, typing.Any] | typing.Any:
        return self.value



import typing

import betterproto2

from betterproto2_compiler.lib.google.protobuf import (
    ListValue as VanillaListValue,
    NullValue,
    Struct as VanillaStruct,
    Value as VanillaValue,
)


class Struct(VanillaStruct):
    # TODO typing
    @classmethod
    def from_dict(cls, value, *, ignore_unknown_fields: bool = False):
        assert isinstance(value, dict)

        fields: dict[str, Value] = {}

        for key, val in value.items():
            fields[key] = Value.from_dict(val)

        return cls(fields=fields)  # type: ignore[reportArgumentType]

    # TODO typing
    def to_dict(
        self,
        *,
        output_format: betterproto2.OutputFormat = betterproto2.OutputFormat.PROTO_JSON,
        casing: betterproto2.Casing = betterproto2.Casing.CAMEL,
        include_default_values: bool = False,
    ) -> dict[str, typing.Any] | typing.Any:
        # If the output format is PYTHON, we should have kept the wrapped type without building the real class
        assert output_format == betterproto2.OutputFormat.PROTO_JSON

        return {
            key: value.to_dict(
                output_format=output_format, casing=casing, include_default_values=include_default_values
            )
            for key, value in self.fields.items()
        }


# We can't use the unwrap mechanism to support directly using a Python object due to the None case: it would then be
# impossible to distinguish between the absence of the message and a None value.
class Value(VanillaValue):
    # TODO typing
    @classmethod
    def from_dict(cls, value, *, ignore_unknown_fields: bool = False):
        match value:
            case bool() as b:
                return cls(bool_value=b)
            case int() | float() as num:
                return cls(number_value=num)
            case str() as s:
                return cls(string_value=s)
            case list() as l:
                return cls(list_value=ListValue.from_dict(l))
            case dict() as d:
                return cls(struct_value=Struct.from_dict(d))
            case None:
                return cls(null_value=NullValue.NULL_VALUE)
        raise ValueError(f"Unknown value type: {type(value)}")

    # TODO typing
    def to_dict(
        self,
        *,
        output_format: betterproto2.OutputFormat = betterproto2.OutputFormat.PROTO_JSON,
        casing: betterproto2.Casing = betterproto2.Casing.CAMEL,
        include_default_values: bool = False,
    ) -> dict[str, typing.Any] | typing.Any:
        # If the output format is PYTHON, we should have kept the wrapped type without building the real class
        assert output_format == betterproto2.OutputFormat.PROTO_JSON

        match self:
            case Value(null_value=NullValue.NULL_VALUE):
                return None
            case Value(bool_value=bool(b)):
                return b
            case Value(number_value=int(num)) | Value(number_value=float(num)):
                return num
            case Value(string_value=str(s)):
                return s
            case Value(list_value=ListValue(values=l)):
                return [v.to_dict() for v in l]
            case Value(struct_value=Struct(fields=f)):
                return {k: v.to_dict() for k, v in f.items()}

        raise ValueError("Invalid value")


class ListValue(VanillaListValue):
    # TODO typing
    @classmethod
    def from_dict(cls, value, *, ignore_unknown_fields: bool = False):
        return cls(values=[Value.from_dict(v) for v in value])

    # TODO typing
    def to_dict(
        self,
        *,
        output_format: betterproto2.OutputFormat = betterproto2.OutputFormat.PROTO_JSON,
        casing: betterproto2.Casing = betterproto2.Casing.CAMEL,
        include_default_values: bool = False,
    ) -> dict[str, typing.Any] | typing.Any:
        # If the output format is PYTHON, we should have kept the wrapped type without building the real class
        assert output_format == betterproto2.OutputFormat.PROTO_JSON

        return [value.to_dict() for value in self.values]



import pathlib
import sys
from collections.abc import Generator

from betterproto2_compiler.lib.google.protobuf import (
    DescriptorProto,
    EnumDescriptorProto,
    FileDescriptorProto,
    ServiceDescriptorProto,
)
from betterproto2_compiler.lib.google.protobuf.compiler import (
    CodeGeneratorRequest,
    CodeGeneratorResponse,
    CodeGeneratorResponseFeature,
    CodeGeneratorResponseFile,
)
from betterproto2_compiler.settings import ClientGeneration, ServerGeneration, Settings

from .compiler import outputfile_compiler
from .models import (
    EnumDefinitionCompiler,
    FieldCompiler,
    MapEntryCompiler,
    MessageCompiler,
    OneofCompiler,
    OneOfFieldCompiler,
    OutputTemplate,
    PluginRequestCompiler,
    ServiceCompiler,
    ServiceMethodCompiler,
    is_map,
    is_oneof,
)


def traverse(
    proto_file: FileDescriptorProto,
) -> Generator[tuple[EnumDescriptorProto | DescriptorProto, list[int], str], None, None]:
    # Todo: Keep information about nested hierarchy
    def _traverse(
        path: list[int],
        items: list[EnumDescriptorProto] | list[DescriptorProto],
        prefix: str = "",
    ) -> Generator[tuple[EnumDescriptorProto | DescriptorProto, list[int], str], None, None]:
        for i, item in enumerate(items):
            # Adjust the name since we flatten the hierarchy.
            should_rename = not isinstance(item, DescriptorProto) or not item.options or not item.options.map_entry

            # Record prefixed name but *do not* mutate original file.
            # We use this prefixed name to create pythonized names.
            prefixed_name = next_prefix = f"{prefix}.{item.name}" if prefix and should_rename else item.name
            yield item, [*path, i], prefixed_name

            if isinstance(item, DescriptorProto):
                # Get nested types.
                yield from _traverse([*path, i, 4], item.enum_type, next_prefix)
                yield from _traverse([*path, i, 3], item.nested_type, next_prefix)

    yield from _traverse([5], proto_file.enum_type)
    yield from _traverse([4], proto_file.message_type)


def get_settings(plugin_options: list[str]) -> Settings:
    # Synchronous clients are suitable for most users
    client_generation = ClientGeneration.SYNC
    server_generation = ServerGeneration.NONE

    for opt in plugin_options:
        if opt.startswith("client_generation="):
            name = opt.split("=")[1]
            try:
                client_generation = ClientGeneration(name)
            except ValueError:
                raise ValueError(f"Invalid client_generation option: {name}")

        if opt.startswith("server_generation="):
            name = opt.split("=")[1]
            try:
                server_generation = ServerGeneration(name)
            except ValueError:
                raise ValueError(f"Invalid server_generation option: {name}")

    return Settings(
        pydantic_dataclasses="pydantic_dataclasses" in plugin_options,
        google_protobuf_descriptors="google_protobuf_descriptors" in plugin_options,
        client_generation=client_generation,
        server_generation=server_generation,
    )


def generate_code(request: CodeGeneratorRequest) -> CodeGeneratorResponse:
    response = CodeGeneratorResponse(supported_features=CodeGeneratorResponseFeature.PROTO3_OPTIONAL)

    plugin_options = request.parameter.split(",") if request.parameter else []
    settings = get_settings(plugin_options)

    request_data = PluginRequestCompiler(plugin_request_obj=request)
    # Gather output packages
    for proto_file in request.proto_file:
        output_package_name = proto_file.package
        if output_package_name not in request_data.output_packages:
            # Create a new output if there is no output for this package
            request_data.output_packages[output_package_name] = OutputTemplate(
                parent_request=request_data, package_proto_obj=proto_file, settings=settings
            )
        # Add this input file to the output corresponding to this package
        request_data.output_packages[output_package_name].input_files.append(proto_file)

    # Read Messages and Enums
    # We need to read Messages before Services in so that we can
    # get the references to input/output messages for each service
    for output_package_name, output_package in request_data.output_packages.items():
        for proto_input_file in output_package.input_files:
            for item, path, prefixed_proto_name in traverse(proto_input_file):
                read_protobuf_type(
                    source_file=proto_input_file,
                    item=item,
                    path=path,
                    output_package=output_package,
                    prefixed_proto_name=prefixed_proto_name,
                )

    # Read Services
    for output_package_name, output_package in request_data.output_packages.items():
        for proto_input_file in output_package.input_files:
            for index, service in enumerate(proto_input_file.service):
                read_protobuf_service(proto_input_file, service, index, output_package)

    # All the hierarchy is ready. We can perform pre-computations before generating the output files
    for package in request_data.output_packages.values():
        for message in package.messages.values():
            for field in message.fields:
                field.ready()
            message.ready()
        for enum in package.enums.values():
            enum.ready()
        for service in package.services.values():
            for method in service.methods:
                method.ready()
            service.ready()

    # Generate output files
    output_paths: set[pathlib.Path] = set()
    for output_package_name, output_package in request_data.output_packages.items():
        # Add files to the response object
        output_path = pathlib.Path(*output_package_name.split("."), "__init__.py")
        output_paths.add(output_path)

        response.file.append(
            CodeGeneratorResponseFile(
                name=str(output_path),
                # Render and then format the output file
                content=outputfile_compiler(output_file=output_package),
            ),
        )

    # Make each output directory a package with __init__ file
    init_files = {
        directory.joinpath("__init__.py")
        for path in output_paths
        for directory in path.parents
        if not directory.joinpath("__init__.py").exists()
    } - output_paths

    for init_file in init_files:
        response.file.append(CodeGeneratorResponseFile(name=str(init_file)))

    response.file.append(
        CodeGeneratorResponseFile(
            name="message_pool.py", content="import betterproto2\n\ndefault_message_pool = betterproto2.MessagePool()\n"
        )
    )

    response.file.append(CodeGeneratorResponseFile(name="py.typed", content=""))

    if settings.google_protobuf_descriptors:
        response.file.append(
            CodeGeneratorResponseFile(
                name="google_proto_descriptor_pool.py",
                content="from google.protobuf import descriptor_pool\n\n"
                + "default_google_proto_descriptor_pool = descriptor_pool.DescriptorPool()\n",
            )
        )

    for output_package_name in sorted(output_paths.union(init_files)):
        print(f"Writing {output_package_name}", file=sys.stderr)

    return response


def read_protobuf_type(
    item: DescriptorProto | EnumDescriptorProto,
    path: list[int],
    source_file: "FileDescriptorProto",
    output_package: OutputTemplate,
    prefixed_proto_name: str,
) -> None:
    if isinstance(item, DescriptorProto):
        if item.options and item.options.map_entry:
            # Skip generated map entry messages since we just use dicts
            return
        # Process Message
        message_data = MessageCompiler(
            source_file=source_file,
            output_file=output_package,
            prefixed_proto_name=prefixed_proto_name,
            proto_obj=item,
            path=path,
        )
        output_package.messages[message_data.prefixed_proto_name] = message_data

        for index, field in enumerate(item.field):
            if is_map(field, item):
                message_data.fields.append(
                    MapEntryCompiler(
                        source_file=source_file,
                        message=message_data,
                        proto_obj=field,
                        path=path + [2, index],
                    )
                )
            elif is_oneof(field):
                message_data.fields.append(
                    OneOfFieldCompiler(
                        source_file=source_file,
                        message=message_data,
                        proto_obj=field,
                        path=path + [2, index],
                    )
                )
            else:
                message_data.fields.append(
                    FieldCompiler(
                        source_file=source_file,
                        message=message_data,
                        proto_obj=field,
                        path=path + [2, index],
                    )
                )

        # Synthetic oneofs are generated for optional fields. We should not generate code or documentation for them.
        is_synthetic_oneof = [True for _ in item.oneof_decl]
        for field in item.field:
            if field.oneof_index is not None and not field.proto3_optional:
                is_synthetic_oneof[field.oneof_index] = False

        for index, (oneof, is_synthetic) in enumerate(zip(item.oneof_decl, is_synthetic_oneof)):
            if not is_synthetic:
                message_data.oneofs.append(
                    OneofCompiler(
                        source_file=source_file,
                        path=path + [8, index],
                        proto_obj=oneof,
                    )
                )

    elif isinstance(item, EnumDescriptorProto):
        # Enum
        enum = EnumDefinitionCompiler(
            source_file=source_file,
            output_file=output_package,
            prefixed_proto_name=prefixed_proto_name,
            proto_obj=item,
            path=path,
        )
        output_package.enums[enum.prefixed_proto_name] = enum


def read_protobuf_service(
    source_file: FileDescriptorProto,
    service: ServiceDescriptorProto,
    index: int,
    output_package: OutputTemplate,
) -> None:
    service_data = ServiceCompiler(
        source_file=source_file,
        output_file=output_package,
        proto_obj=service,
        path=[6, index],
    )
    service_data.output_file.services[service_data.proto_name] = service_data

    for j, method in enumerate(service.method):
        service_data.methods.append(
            ServiceMethodCompiler(
                source_file=source_file,
                parent=service_data,
                proto_obj=method,
                path=[6, index, 2, j],
            )
        )



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: protobuf.proto
# plugin: python-betterproto2
# This file has been @generated

__all__ = ()


import betterproto2

_COMPILER_VERSION = "0.9.0"
betterproto2.check_compiler_version(_COMPILER_VERSION)



import keyword
import re

# Word delimiters and symbols that will not be preserved when re-casing.
# language=PythonRegExp
SYMBOLS = "[^a-zA-Z0-9]*"

# Optionally capitalized word.
# language=PythonRegExp
WORD = "[A-Z]*[a-z]*[0-9]*"

# Uppercase word, not followed by lowercase letters.
# language=PythonRegExp
WORD_UPPER = "[A-Z]+(?![a-z])[0-9]*"


def safe_snake_case(value: str) -> str:
    """Snake case a value taking into account Python keywords."""
    value = snake_case(value)
    value = sanitize_name(value)
    return value


def snake_case(name: str) -> str:
    """
    Join words with an underscore into lowercase and remove symbols.
    """

    # If there are already underscores in the name, don't break it
    if "_" in name or not any([c.isupper() for c in name]):
        return name

    # Add an underscore before capital letters
    name = re.sub(r"(?<=[a-z0-9])([A-Z])", r"_\1", name)

    # Add an underscore before capital letters following an acronym
    name = re.sub(r"(?<=[A-Z])([A-Z])(?=[a-z])", r"_\1", name)

    # Add an underscore before digits
    name = re.sub(r"(?<=[a-zA-Z])([0-9])", r"_\1", name)

    return name.lower()


def pascal_case(value: str, strict: bool = True) -> str:
    """
    Capitalize each word and remove symbols.

    Parameters
    -----------
    value: :class:`str`
        The value to convert.
    strict: :class:`bool`
        Whether or not to output only alphanumeric characters.

    Returns
    --------
    :class:`str`
        The value in PascalCase.
    """

    def substitute_word(symbols, word):
        if strict:
            return word.capitalize()  # Remove all delimiters

        if word.islower():
            delimiter_length = len(symbols[:-1])  # Lose one delimiter
        else:
            delimiter_length = len(symbols)  # Preserve all delimiters

        return ("_" * delimiter_length) + word.capitalize()

    return re.sub(
        f"({SYMBOLS})({WORD_UPPER}|{WORD})",
        lambda groups: substitute_word(groups[1], groups[2]),
        value,
    )


def camel_case(value: str, strict: bool = True) -> str:
    """
    Capitalize all words except first and remove symbols.

    Parameters
    -----------
    value: :class:`str`
        The value to convert.
    strict: :class:`bool`
        Whether or not to output only alphanumeric characters.

    Returns
    --------
    :class:`str`
        The value in camelCase.
    """
    return lowercase_first(pascal_case(value, strict=strict))


def lowercase_first(value: str) -> str:
    """
    Lower cases the first character of the value.

    Parameters
    ----------
    value: :class:`str`
        The value to lower case.

    Returns
    -------
    :class:`str`
        The lower cased string.
    """
    return value[0:1].lower() + value[1:]


def sanitize_name(value: str) -> str:
    # https://www.python.org/dev/peps/pep-0008/#descriptive-naming-styles
    if keyword.iskeyword(value):
        return f"{value}_"
    if not value.isidentifier():
        return f"_{value}"
    return value



#!/usr/bin/env python

import os
import sys

from betterproto2_compiler.lib.google.protobuf.compiler import (
    CodeGeneratorRequest,
)
from betterproto2_compiler.plugin.parser import generate_code


def main() -> None:
    """The plugin's main entry point."""
    # Read request message from stdin
    data = sys.stdin.buffer.read()

    # Parse request
    request = CodeGeneratorRequest.parse(data)

    dump_file = os.getenv("BETTERPROTO_DUMP")
    if dump_file:
        dump_request(dump_file, request)

    # Generate code
    response = generate_code(request)

    # Serialise response message
    output = response.SerializeToString()

    # Write to stdout
    sys.stdout.buffer.write(output)


def dump_request(dump_file: str, request: CodeGeneratorRequest) -> None:
    """
    For developers: Supports running plugin.py standalone so its possible to debug it.
    Run protoc (or generate.py) with BETTERPROTO_DUMP="yourfile.bin" to write the request to a file.
    Then run plugin.py from your IDE in debugging mode, and redirect stdin to the file.
    """
    with open(str(dump_file), "wb") as fh:
        sys.stderr.write(f"\033[31mWriting input from protoc to: {dump_file}\033[0m\n")
        fh.write(request.SerializeToString())


if __name__ == "__main__":
    main()



def test_snake_case() -> None:
    from betterproto2_compiler.casing import snake_case

    # Simple renaming
    assert snake_case("methodName") == "method_name"
    assert snake_case("MethodName") == "method_name"

    # Don't break acronyms
    assert snake_case("HTTPRequest") == "http_request"
    assert snake_case("RequestHTTP") == "request_http"
    assert snake_case("HTTPRequest2") == "http_request_2"
    assert snake_case("RequestHTTP2") == "request_http_2"
    assert snake_case("GetAResponse") == "get_a_response"

    # Split digits
    assert snake_case("Get2025Results") == "get_2025_results"
    assert snake_case("Get10yResults") == "get_10y_results"

    # If the name already contains an underscore or is lowercase, don't change it at all.
    # There is a risk of breaking names otherwise.
    assert snake_case("aaa_123_bbb") == "aaa_123_bbb"
    assert snake_case("aaa_123bbb") == "aaa_123bbb"
    assert snake_case("aaa123_bbb") == "aaa123_bbb"
    assert snake_case("get_HTTP_response") == "get_HTTP_response"
    assert snake_case("_methodName") == "_methodName"
    assert snake_case("make_gRPC_request") == "make_gRPC_request"

    assert snake_case("value1") == "value1"
    assert snake_case("value1string") == "value1string"

    # It is difficult to cover all the cases with a simple algorithm...
    # "GetValueAsUInt32" -> "get_value_as_u_int_32"



import os.path
import subprocess
import sys
from importlib import metadata

from .module_validation import ModuleValidator

try:
    # betterproto[compiler] specific dependencies
    import jinja2
except ImportError as err:
    print(
        "\033[31m"
        f"Unable to import `{err.name}` from betterproto plugin! "
        "Please ensure that you've installed betterproto as "
        '`pip install "betterproto[compiler]"` so that compiler dependencies '
        "are included."
        "\033[0m",
    )
    raise SystemExit(1)

from .models import OutputTemplate


def outputfile_compiler(output_file: OutputTemplate) -> str:
    templates_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "templates"))

    version = metadata.version("betterproto2_compiler")

    env = jinja2.Environment(
        trim_blocks=True,
        lstrip_blocks=True,
        loader=jinja2.FileSystemLoader(templates_folder),
        undefined=jinja2.StrictUndefined,
    )

    # List of the symbols that should appear in the `__all__` variable of the file
    all: list[str] = []

    def add_to_all(name: str) -> str:
        all.append(name)
        return name

    env.filters["add_to_all"] = add_to_all

    body_template = env.get_template("template.py.j2")
    header_template = env.get_template("header.py.j2")

    # Load the body first do know the symbols defined in the file
    code = body_template.render(output_file=output_file)
    code = header_template.render(output_file=output_file, version=version, all=all) + "\n" + code

    try:
        # Sort imports, delete unused ones, sort __all__
        code = subprocess.check_output(
            ["ruff", "check", "--select", "I,F401,TC005,RUF022", "--fix", "--silent", "-"],
            input=code,
            encoding="utf-8",
        )

        # Format the code
        code = subprocess.check_output(["ruff", "format", "-"], input=code, encoding="utf-8")
    except subprocess.CalledProcessError:
        with open("invalid-generated-code.py", "w") as f:
            f.write(code)

        raise SyntaxError(
            f"Can't format the source code:\nThe invalid generated code has been written in `invalid-generated-code.py`"
        )

    # Validate the generated code.
    validator = ModuleValidator(iter(code.splitlines()))
    if not validator.validate():
        message_builder = ["[WARNING]: Generated code has collisions in the module:"]
        for collision, lines in validator.collisions.items():
            message_builder.append(f'  "{collision}" on lines:')
            for num, line in lines:
                message_builder.append(f"    {num}:{line}")
        print("\n".join(message_builder), file=sys.stderr)
    return code



#!/usr/bin/env python
import asyncio
import os
import shutil

from tests.util import protoc

# Force pure-python implementation instead of C++, otherwise imports
# break things because we can't properly reset the symbol database.
os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"


async def generate_test(
    name,
    semaphore: asyncio.Semaphore,
    *,
    reference: bool = False,
    pydantic: bool = False,
    descriptors: bool = False,
    client_generation: str = "async_sync",
):
    await semaphore.acquire()

    dir_path = os.path.dirname(os.path.realpath(__file__))

    options = []
    if reference:
        options.append("reference")
    if pydantic:
        options.append("pydantic")
    if descriptors:
        options.append("descriptors")

    input_dir = dir_path + "/inputs/" + name
    output_dir = dir_path + "/outputs/" + name + ("_" + "_".join(options) if options else "")

    os.mkdir(output_dir)

    stdout, stderr, returncode = await protoc(
        input_dir,
        output_dir,
        reference=reference,
        pydantic_dataclasses=pydantic,
        google_protobuf_descriptors=descriptors,
        client_generation=client_generation,
    )

    if options:
        options_str = ", ".join(options)
        options_str = f" ({options_str})"
    else:
        options_str = ""

    if returncode == 0:
        print(f"\033[31;1;4mGenerated output for {name!r}{options_str}\033[0m")
    else:
        print(f"\033[31;1;4mFailed to generate reference output for {name!r}{options_str}\033[0m")
        print(stderr.decode())

    semaphore.release()


async def main_async():
    # Don't compile too many tests in parallel
    semaphore = asyncio.Semaphore(os.cpu_count() or 1)

    tasks = [
        generate_test("any", semaphore),
        generate_test("bool", semaphore, pydantic=True),
        generate_test("bool", semaphore, reference=True),
        generate_test("bool", semaphore),
        generate_test("bytes", semaphore, reference=True),
        generate_test("bytes", semaphore),
        generate_test("casing_inner_class", semaphore),
        generate_test("casing", semaphore, reference=True),
        generate_test("casing", semaphore),
        generate_test("compiler_lib", semaphore),
        generate_test("deprecated", semaphore, reference=True),
        generate_test("deprecated", semaphore, client_generation="async"),
        generate_test("documentation", semaphore, client_generation="async"),
        generate_test("double", semaphore, reference=True),
        generate_test("double", semaphore),
        generate_test("encoding_decoding", semaphore),
        generate_test("enum", semaphore, reference=True),
        generate_test("enum", semaphore),
        generate_test("example_service", semaphore, client_generation="async"),
        generate_test("features", semaphore),
        generate_test("field_name_identical_to_type", semaphore, reference=True),
        generate_test("field_name_identical_to_type", semaphore),
        generate_test("fixed", semaphore, reference=True),
        generate_test("fixed", semaphore),
        generate_test("float", semaphore, reference=True),
        generate_test("float", semaphore),
        generate_test("google_impl_behavior_equivalence", semaphore, reference=True),
        generate_test("google_impl_behavior_equivalence", semaphore),
        generate_test("google", semaphore),
        generate_test("googletypes_request", semaphore, client_generation="async"),
        generate_test("googletypes_response_embedded", semaphore, client_generation="async"),
        generate_test("googletypes_response", semaphore, client_generation="async"),
        generate_test("googletypes_struct", semaphore, reference=True),
        generate_test("googletypes_struct", semaphore),
        generate_test("googletypes_value", semaphore, reference=True),
        generate_test("googletypes_value", semaphore),
        generate_test("googletypes", semaphore, reference=True),
        generate_test("googletypes", semaphore),
        generate_test("grpclib_reflection", semaphore, descriptors=True, client_generation="async"),
        generate_test("grpclib_reflection", semaphore, client_generation="async"),
        generate_test("import_cousin_package_same_name", semaphore, descriptors=True),
        generate_test("import_cousin_package_same_name", semaphore),
        generate_test("import_service_input_message", semaphore, client_generation="async"),
        generate_test("int32", semaphore, reference=True),
        generate_test("int32", semaphore),
        generate_test("invalid_field", semaphore, pydantic=True),
        generate_test("invalid_field", semaphore),
        generate_test("manual_validation", semaphore, pydantic=True),
        generate_test("manual_validation", semaphore),
        generate_test("map", semaphore, reference=True),
        generate_test("map", semaphore),
        generate_test("mapmessage", semaphore, reference=True),
        generate_test("mapmessage", semaphore),
        generate_test("message_wrapping", semaphore),
        generate_test("namespace_builtin_types", semaphore, reference=True),
        generate_test("namespace_builtin_types", semaphore),
        generate_test("namespace_keywords", semaphore, reference=True),
        generate_test("namespace_keywords", semaphore),
        generate_test("nested", semaphore, reference=True),
        generate_test("nested", semaphore),
        generate_test("nestedtwice", semaphore, reference=True),
        generate_test("nestedtwice", semaphore),
        generate_test("oneof_default_value_serialization", semaphore),
        generate_test("oneof_empty", semaphore, reference=True),
        generate_test("oneof_empty", semaphore),
        generate_test("oneof_enum", semaphore, reference=True),
        generate_test("oneof_enum", semaphore),
        generate_test("oneof", semaphore, pydantic=True),
        generate_test("oneof", semaphore, reference=True),
        generate_test("oneof", semaphore),
        generate_test("pickling", semaphore),
        generate_test("proto3_field_presence_oneof", semaphore, reference=True),
        generate_test("proto3_field_presence_oneof", semaphore),
        generate_test("proto3_field_presence", semaphore, reference=True),
        generate_test("proto3_field_presence", semaphore),
        generate_test("recursivemessage", semaphore, reference=True),
        generate_test("recursivemessage", semaphore),
        generate_test("ref", semaphore, reference=True),
        generate_test("ref", semaphore),
        generate_test("regression_387", semaphore),
        generate_test("regression_414", semaphore),
        generate_test("repeated_duration_timestamp", semaphore, reference=True),
        generate_test("repeated_duration_timestamp", semaphore),
        generate_test("repeated", semaphore, reference=True),
        generate_test("repeated", semaphore),
        generate_test("repeatedmessage", semaphore, reference=True),
        generate_test("repeatedmessage", semaphore),
        generate_test("repeatedpacked", semaphore, reference=True),
        generate_test("repeatedpacked", semaphore),
        generate_test("rpc_empty_input_message", semaphore, client_generation="async"),
        generate_test("service_uppercase", semaphore, client_generation="async"),
        generate_test("service", semaphore),
        generate_test("signed", semaphore, reference=True),
        generate_test("signed", semaphore),
        generate_test("simple_service", semaphore),
        generate_test("stream_stream", semaphore),
        generate_test("timestamp_dict_encode", semaphore, reference=True),
        generate_test("timestamp_dict_encode", semaphore),
        generate_test("unwrap", semaphore),
        generate_test("validation", semaphore, pydantic=True),
    ]
    await asyncio.gather(*tasks)


def main():
    # Clean the output directory
    dir_path = os.path.dirname(os.path.realpath(__file__))

    shutil.rmtree(dir_path + "/outputs", ignore_errors=True)
    os.mkdir(dir_path + "/outputs")

    asyncio.run(main_async())


if __name__ == "__main__":
    main()



import sys
from dataclasses import dataclass

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from strenum import StrEnum


class ClientGeneration(StrEnum):
    NONE = "none"
    """Clients are not generated."""

    SYNC = "sync"
    """Only synchronous clients are generated."""

    ASYNC = "async"
    """Only asynchronous clients are generated."""

    SYNC_ASYNC = "sync_async"
    """Both synchronous and asynchronous clients are generated.

    Asynchronous clients are generated with the Async suffix."""

    ASYNC_SYNC = "async_sync"
    """Both synchronous and asynchronous clients are generated.

    Synchronous clients are generated with the Sync suffix."""

    SYNC_ASYNC_NO_DEFAULT = "sync_async_no_default"
    """Both synchronous and asynchronous clients are generated.

    Synchronous clients are generated with the Sync suffix, and asynchronous clients are generated with the Async
    suffix."""

    @property
    def is_sync_generated(self) -> bool:
        return self in {
            ClientGeneration.SYNC,
            ClientGeneration.SYNC_ASYNC,
            ClientGeneration.ASYNC_SYNC,
            ClientGeneration.SYNC_ASYNC_NO_DEFAULT,
        }

    @property
    def is_async_generated(self) -> bool:
        return self in {
            ClientGeneration.ASYNC,
            ClientGeneration.SYNC_ASYNC,
            ClientGeneration.ASYNC_SYNC,
            ClientGeneration.SYNC_ASYNC_NO_DEFAULT,
        }

    @property
    def is_sync_prefixed(self) -> bool:
        return self in {ClientGeneration.ASYNC_SYNC, ClientGeneration.SYNC_ASYNC_NO_DEFAULT}

    @property
    def is_async_prefixed(self) -> bool:
        return self in {ClientGeneration.SYNC_ASYNC, ClientGeneration.SYNC_ASYNC_NO_DEFAULT}


class ServerGeneration(StrEnum):
    NONE = "none"
    ASYNC = "async"


@dataclass
class Settings:
    pydantic_dataclasses: bool
    google_protobuf_descriptors: bool

    client_generation: ClientGeneration
    server_generation: ServerGeneration



import asyncio
import os
import sys
from pathlib import Path

os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"


async def protoc(
    path: str | Path,
    output_dir: str | Path,
    reference: bool = False,
    pydantic_dataclasses: bool = False,
    google_protobuf_descriptors: bool = False,
    client_generation: str = "async_sync",
):
    resolved_path: Path = Path(path).resolve()
    resolved_output_dir: Path = Path(output_dir).resolve()
    python_out_option: str = "python_out" if reference else "python_betterproto2_out"

    command = [
        sys.executable,
        "-m",
        "grpc.tools.protoc",
        f"--proto_path={resolved_path.as_posix()}",
        f"--{python_out_option}={resolved_output_dir.as_posix()}",
        *[p.as_posix() for p in resolved_path.glob("*.proto")],
    ]

    if not reference:
        command.insert(3, "--python_betterproto2_opt=server_generation=async")
        command.insert(3, f"--python_betterproto2_opt=client_generation={client_generation}")

        if pydantic_dataclasses:
            command.insert(3, "--python_betterproto2_opt=pydantic_dataclasses")

        if google_protobuf_descriptors:
            command.insert(3, "--python_betterproto2_opt=google_protobuf_descriptors")

    proc = await asyncio.create_subprocess_exec(
        *command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await proc.communicate()
    return stdout, stderr, proc.returncode or 0



from collections.abc import Callable

from .any import Any
from .duration import Duration
from .google_values import (
    BoolValue,
    BytesValue,
    DoubleValue,
    FloatValue,
    Int32Value,
    Int64Value,
    StringValue,
    UInt32Value,
    UInt64Value,
)
from .struct import ListValue, Struct, Value
from .timestamp import Timestamp

# For each (package, message name), lists the methods that should be added to the message definition.
# The source code of the method is read from the `known_types` folder. If imports are needed, they can be directly added
# to the template file: they will automatically be removed if not necessary.
KNOWN_METHODS: dict[tuple[str, str], list[Callable]] = {
    ("google.protobuf", "Any"): [Any.pack, Any.unpack, Any.to_dict, Any.from_dict],
    ("google.protobuf", "Timestamp"): [
        Timestamp.from_datetime,
        Timestamp.to_datetime,
        Timestamp.timestamp_to_json,
        Timestamp.from_dict,
        Timestamp.to_dict,
        Timestamp.from_wrapped,
        Timestamp.to_wrapped,
    ],
    ("google.protobuf", "Duration"): [
        Duration.from_timedelta,
        Duration.to_timedelta,
        Duration.delta_to_json,
        Duration.from_dict,
        Duration.to_dict,
        Duration.from_wrapped,
        Duration.to_wrapped,
    ],
    ("google.protobuf", "BoolValue"): [
        BoolValue.from_dict,
        BoolValue.to_dict,
        BoolValue.from_wrapped,
        BoolValue.to_wrapped,
    ],
    ("google.protobuf", "Int32Value"): [
        Int32Value.from_dict,
        Int32Value.to_dict,
        Int32Value.from_wrapped,
        Int32Value.to_wrapped,
    ],
    ("google.protobuf", "Int64Value"): [
        Int64Value.from_dict,
        Int64Value.to_dict,
        Int64Value.from_wrapped,
        Int64Value.to_wrapped,
    ],
    ("google.protobuf", "UInt32Value"): [
        UInt32Value.from_dict,
        UInt32Value.to_dict,
        UInt32Value.from_wrapped,
        UInt32Value.to_wrapped,
    ],
    ("google.protobuf", "UInt64Value"): [
        UInt64Value.from_dict,
        UInt64Value.to_dict,
        UInt64Value.from_wrapped,
        UInt64Value.to_wrapped,
    ],
    ("google.protobuf", "FloatValue"): [
        FloatValue.from_dict,
        FloatValue.to_dict,
        FloatValue.from_wrapped,
        FloatValue.to_wrapped,
    ],
    ("google.protobuf", "DoubleValue"): [
        DoubleValue.from_dict,
        DoubleValue.to_dict,
        DoubleValue.from_wrapped,
        DoubleValue.to_wrapped,
    ],
    ("google.protobuf", "StringValue"): [
        StringValue.from_dict,
        StringValue.to_dict,
        StringValue.from_wrapped,
        StringValue.to_wrapped,
    ],
    ("google.protobuf", "BytesValue"): [
        BytesValue.from_dict,
        BytesValue.to_dict,
        BytesValue.from_wrapped,
        BytesValue.to_wrapped,
    ],
    ("google.protobuf", "Struct"): [
        Struct.from_dict,
        Struct.to_dict,
    ],
    ("google.protobuf", "ListValue"): [
        ListValue.from_dict,
        ListValue.to_dict,
    ],
    ("google.protobuf", "Value"): [
        Value.from_dict,
        Value.to_dict,
    ],
}

# A wrapped type is the type of a message that is automatically replaced by a known Python type.
WRAPPED_TYPES: dict[tuple[str, str], str] = {
    ("google.protobuf", "BoolValue"): "bool",
    ("google.protobuf", "Int32Value"): "int",
    ("google.protobuf", "Int64Value"): "int",
    ("google.protobuf", "UInt32Value"): "int",
    ("google.protobuf", "UInt64Value"): "int",
    ("google.protobuf", "FloatValue"): "float",
    ("google.protobuf", "DoubleValue"): "float",
    ("google.protobuf", "StringValue"): "str",
    ("google.protobuf", "BytesValue"): "bytes",
    ("google.protobuf", "Timestamp"): "datetime.datetime",
    ("google.protobuf", "Duration"): "datetime.timedelta",
}



import pytest

from betterproto2_compiler.plugin.module_validation import ModuleValidator


@pytest.mark.parametrize(
    ["text", "expected_collisions"],
    [
        pytest.param(
            ["import os"],
            None,
            id="single import",
        ),
        pytest.param(
            ["import os", "import sys"],
            None,
            id="multiple imports",
        ),
        pytest.param(
            ["import os", "import os"],
            {"os"},
            id="duplicate imports",
        ),
        pytest.param(
            ["from os import path", "import os"],
            None,
            id="duplicate imports with alias",
        ),
        pytest.param(
            ["from os import path", "import os as os_alias"],
            None,
            id="duplicate imports with alias",
        ),
        pytest.param(
            ["from os import path", "import os as path"],
            {"path"},
            id="duplicate imports with alias",
        ),
        pytest.param(
            ["import os", "class os:"],
            {"os"},
            id="duplicate import with class",
        ),
        pytest.param(
            ["import os", "class os:", "  pass", "import sys"],
            {"os"},
            id="duplicate import with class and another",
        ),
        pytest.param(
            ["def test(): pass", "class test:"],
            {"test"},
            id="duplicate class and function",
        ),
        pytest.param(
            ["def test(): pass", "def test(): pass"],
            {"test"},
            id="duplicate functions",
        ),
        pytest.param(
            ["def test(): pass", "test = 100"],
            {"test"},
            id="function and variable",
        ),
        pytest.param(
            ["def test():", "    test = 3"],
            None,
            id="function and variable in function",
        ),
        pytest.param(
            [
                "def test(): pass",
                "'''",
                "def test(): pass",
                "'''",
                "def test_2(): pass",
            ],
            None,
            id="duplicate functions with multiline string",
        ),
        pytest.param(
            ["def test(): pass", "# def test(): pass"],
            None,
            id="duplicate functions with comments",
        ),
        pytest.param(
            ["from test import (", "    A", "   B", "   C", ")"],
            None,
            id="multiline import",
        ),
        pytest.param(
            ["from test import (", "    A", "   B", "   C", ")", "from test import A"],
            {"A"},
            id="multiline import with duplicate",
        ),
    ],
)
def test_module_validator(text: list[str], expected_collisions: set[str] | None):
    line_iterator = iter(text)
    validator = ModuleValidator(line_iterator)
    valid = validator.validate()
    if expected_collisions is None:
        assert valid
    else:
        assert set(validator.collisions.keys()) == expected_collisions
        assert not valid
